#include <stdio.h>
#include <string.h>
#include <linux/ip.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <lib/util.h>
#include <lib/net/ipextra.h>

/*

    0                   1                   2                   3   
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Version|  IHL  |Type of Service|          Total Length         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |         Identification        |Flags|      Fragment Offset    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  Time to Live |    Protocol   |         Header Checksum       |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                       Source Address                          |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Destination Address                        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Options                    |    Padding    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                     /             \
                    /               \
                   /                 \
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   | Option Fields |    Options Data             ~                 |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    
*/

// assume number of byte in buffer is even !!!
unsigned short ipextra_checksum(unsigned char* buf, int len)
{
  unsigned short *ptr = (unsigned short *) buf;
  int nshort = (len >> 1);
  unsigned long ret;
  int i;

  ret = 0;
  for (i=0; i<nshort; i++) {
    ret += ptr[i];
  }
  
  while ((ret & 0xffff0000) != 0) {
    ret = (ret & 0xffff) + ((ret >> 16) & 0xffff);
  }
  ret = ~ret;

  return (unsigned short) (ret & 0xffff);
}

void ipextra_add_option(char *src, int datalen, struct ipopt *opt, int optlen)
{
  struct iphdr *iph = (struct iphdr *) src;
  struct ipopt *ptr = (struct ipopt *) (src + sizeof(struct iphdr));
  char *cp;
  int len;

/*
       original  -------------->   after

   +-------------+<- iph      +-------------+ 
   |    iphdr    |<----- cp'  |    iphdr    |
   +-------------+<- ptr      +-------------+
   +-------------+            |    ipopt    |
   |      2      |            +-------------+
   +-------------+            +-------------+
   |      1      |       ^    |      2      |
   +-------------+       ^    +-------------+
   |      0      |       ^    |      1      |
   +-------------+<----- cp   +-------------+
                              |      0      |
                              +-------------+
*/
  DB("src=%x payload=%x end=%x\n",
     (int) src,
     (int) (src + sizeof(struct iphdr)),
     (int) (src + sizeof(struct iphdr) + datalen));

  // move IP payload, start from the end with block size = optlen
  cp = src + sizeof(struct iphdr) + datalen;

  len = optlen;  
  do {
    cp -= optlen;
    if (cp < (char *) ptr) {
      len -= (((char *) ptr) - cp);
      cp = (char *) ptr;
    }
    DB(">>>  %x <- %x %d\n", (int) cp+optlen, (int) cp, len);
    memcpy(cp+optlen, cp, len);
  } while (cp > (char *) ptr);
  
  // copy IP option
  memcpy(ptr, opt, optlen);

  // set IP option field
  // ptr->optlen = optlen;
  
  // set IP header
  iph->tot_len = htons(ntohs(iph->tot_len) + optlen);
  iph->ihl += (optlen >> 2);
  iph->check = 0;
  iph->check = ipextra_checksum((unsigned char *) iph, iph->ihl << 2);
}

void ipextra_add_option2(char *dest, const char *src, int datalen, struct ipopt *opt,
                        int optlen)
{
  struct iphdr *iph = (struct iphdr *) dest;
  struct ipopt *ptr = (struct ipopt *) (dest + sizeof(struct iphdr));
  
  // copy IP header
  memcpy(dest, src, sizeof(struct iphdr));

  // copy IP payload
  memcpy(dest + sizeof(struct iphdr) + optlen, 
         src + sizeof(struct iphdr),
         datalen - sizeof(struct iphdr));

  // copy IP option
  memcpy(ptr, opt, optlen);

  // set IP option field
  // ptr->optlen = optlen;
  
  // set IP header
  iph->tot_len = htons(ntohs(iph->tot_len) + optlen);
  iph->ihl += (optlen >> 2);
  iph->check = 0;
  iph->check = ipextra_checksum((unsigned char *) iph, iph->ihl << 2);
}

void ipextra_del_option(char *src, int datalen, int optlen)
{
  struct iphdr *iph = (struct iphdr *) src;
  char *endptr;
  char *cp;
  int len;

/*
        after   <---------------  original
       
   +-------------+<- iph      +-------------+ 
   |    iphdr    |            |    iphdr    |
   +-------------+   cp ----->+-------------+
   |      0      |            |    ipopt    |
   +-------------+    v       +-------------+
   |      1      |    v       |      0      |
   +-------------+    v       +-------------+
   |      2      |            |      1      |
   +-------------+            +-------------+
   +-------------+   endptr ->|      2      |
                     cp'----->+-------------+
                              +-------------+
*/

  DB("src=%x payload=%x end=%x\n",
     (int) src,
     (int) (src + sizeof(struct iphdr)),
     (int) (src + sizeof(struct iphdr) + datalen));

  // move IP payload, start from the top with block size = optlen
  cp = src + sizeof(struct iphdr);
  endptr = cp + datalen - optlen;

  len = optlen;
  do {
    cp += optlen;
    if (cp > endptr) {
      len -= (cp - endptr);
    }
    DB(">>>  %x <- %x %d\n", (int) (cp-optlen), (int) cp, len);
    memcpy(cp-optlen, cp, len);
  } while (cp < endptr);
  
  // set IP header
  iph->tot_len = htons(ntohs(iph->tot_len) - optlen);
  iph->ihl -= (optlen >> 2);
  iph->check = 0;
  iph->check = ipextra_checksum((unsigned char *) iph, iph->ihl << 2);
}

/*
 * dest - destination buffer
 * src  - source buffer
 * datalen - total IP packet size in source buffer
 * optlen - IP option length
 */
void ipextra_del_option2(char *dest, const char *src, int datalen, int optlen)
{
  struct iphdr *iph = (struct iphdr *) dest;
  
  // copy IP header
  memcpy(dest, src, sizeof(struct iphdr));

  // copy IP payload
  memcpy(dest + sizeof(struct iphdr), 
         src + sizeof(struct iphdr) + optlen,
         datalen - sizeof(struct iphdr) - optlen);

  // set IP header
  iph->tot_len = htons(ntohs(iph->tot_len) - optlen);
  iph->ihl -= (optlen >> 2);
  iph->check = 0;
  iph->check = ipextra_checksum((unsigned char *) iph, iph->ihl << 2);
}

int ipextra_getifaddr(char *ifname, struct ifreq *ifr)
{
  int sd;

  if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    return -1;
  }

  memset(ifr, 0, sizeof(struct ifreq));
  strncpy(ifr->ifr_name, ifname, sizeof(ifr->ifr_name));
  if (ioctl(sd, SIOCGIFADDR, (caddr_t) ifr) < 0) {
    close(sd);
    return -1;
  }

  close(sd);
  return 0;
}
