#include <lib/security/auth.h>

static BIO *bio_err;

void auth_init_bio_err()
{
  bio_err = BIO_new_fp(stderr, BIO_NOCLOSE | BIO_FP_TEXT);
}

void auth_free_bio_err()
{
  BIO_free(bio_err);
}

// ----------------------------------------------------------------------------
//
//  digital signature
//
EVP_PKEY *auth_load_privatekey(int filetype, const char *filename,
                               const char *password)
{     
  EVP_PKEY *pkey;
  PKCS12 *p12;
  BIO *bio;

  if ((bio = BIO_new_file(filename, "r")) == NULL) {
    BIO_printf(bio_err, "Error opening key file %s\n", filename);
    ERR_print_errors(bio_err);
    return NULL;
  }

  switch (filetype) {

  case FILETYPE_PKCS12:
    p12 = d2i_PKCS12_bio(bio, NULL);
    BIO_free(bio);

    if (!p12)
      ERR_RET(NULL, "d2i_PKCS12_fp: %s", SSL_ERR_STRING);
    
    if (!PKCS12_parse(p12, password, &pkey, NULL, NULL))
      ERR_RET(NULL, "PKCS12_parse: %s", SSL_ERR_STRING);
    break;

  case FILETYPE_PEM:
    pkey = PEM_read_bio_PrivateKey(bio, NULL, NULL, (char *) password);
    BIO_free(bio);
    break;

  default:
    BIO_free(bio);
    return NULL;

  }

  if (pkey == NULL)
    ERR_RET(NULL, "PEM_ASN1_read: %s", SSL_ERR_STRING);

  return pkey;
}

EVP_PKEY *auth_load_publickey(int filetype, char *filename, char *password)
{
  EVP_PKEY *pkey;
  PKCS12 *p12;
  X509 *x509;
  BIO *bio;

  if ((bio = BIO_new_file(filename, "r")) == NULL) {
    BIO_printf(bio_err, "Error opening certificate file %s\n", filename);
    ERR_print_errors(bio_err);
    return NULL;
  }

  switch (filetype) {

  case FILETYPE_PKCS12:
    p12 = d2i_PKCS12_bio(bio, NULL);
    BIO_free(bio);

    if (!p12)
      ERR_RET(NULL, "d2i_PKCS12_fp: %s", SSL_ERR_STRING);
    if (!PKCS12_parse(p12, password, NULL, &x509, NULL))
      ERR_RET(NULL, "PKCS12_parse: %s", SSL_ERR_STRING);
    break;

  case FILETYPE_PEM:
    x509 = PEM_read_bio_X509(bio, NULL, NULL, NULL);
    BIO_free(bio);
    break;

  default:
    BIO_free(bio);
    return NULL;

  }

  if (x509 == NULL)
    ERR_RET(NULL, "PEM_read_x509: %s", SSL_ERR_STRING);
  
  /* get public key */
  pkey = X509_extract_key(x509);
  if (pkey == NULL)
    ERR_RET(NULL, "X509_extract_key: %s", SSL_ERR_STRING);

  return pkey;
}

int auth_sign_buffer(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *buf,
                     int buflen, char *sigbuf, int *siglen)
{
  EVP_SignInit(ctx, EVP_sha1());
  EVP_SignUpdate(ctx, buf, buflen);
  return EVP_SignFinal(ctx, sigbuf, siglen, pkey);
}

/*
 * returns 1 for a correct signature, 0 for
 * failure and -1 if some other error occurred
 */
int auth_verify_buffer(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *buf,
                       int buflen, char *sigbuf, int siglen)
{
  EVP_VerifyInit(ctx, EVP_sha1());
  EVP_VerifyUpdate(ctx, buf, buflen);
  return EVP_VerifyFinal(ctx, sigbuf, siglen, pkey);
}

/*
 * returns 1 for a correct signature, 0 for
 * failure and -1 if some other error occurred
 */
int auth_sign_file(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *file,
                   int buflen, char *sigbuf, int *siglen)
{
  FILE *fp;
  char *buf;
  int ret, len;

  fp = fopen(file, "r");
  if (fp == NULL)
    return -1;

  buf = malloc(buflen);
  if (buf == NULL) {
    fclose(fp);
    return -1;
  }

  EVP_SignInit(ctx, EVP_sha1());
  while (!feof(fp)) {
    len = fread(buf, 1, buflen, fp);
    EVP_SignUpdate(ctx, buf, len);
  }
  ret = EVP_SignFinal(ctx, sigbuf, siglen, pkey);

  free(buf);
  fclose(fp);

  return ret;
}

int auth_verify_file(EVP_MD_CTX *ctx, EVP_PKEY *pkey, const char *file,
                     int buflen, char *sigbuf, int siglen)
{
  FILE *fp;
  char *buf;
  int ret, len;
  
  fp = fopen(file, "r");
  if (fp == NULL)
    return -1;

  buf = malloc(buflen);
  if (buf == NULL) {
    fclose(fp);
    return -1;
  }

  EVP_VerifyInit(ctx, EVP_sha1());
  while (!feof(fp)) {
    len = fread(buf, 1, buflen, fp);
    EVP_VerifyUpdate(ctx, buf, len);
  }
  ret = EVP_VerifyFinal(ctx, sigbuf, siglen, pkey);

  free(buf);
  fclose(fp);

  return ret;
}

// ----------------------------------------------------------------------------
//
//  certificate verification
//

static int verify_cb(int ok, X509_STORE_CTX *ctx)
{
  /*
    char buf[256];
  */

  if (!ok) {
    /*
      X509_NAME_oneline(X509_get_subject_name(ctx->current_cert), buf, 
      sizeof(buf));
      printf("%s\n", buf);
    */
    printf("error %d at %d depth lookup: %s\n", ctx->error, 
           ctx->error_depth, 
           X509_verify_cert_error_string(ctx->error));

    if ((ctx->error == X509_V_ERR_CERT_HAS_EXPIRED) ||
        (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ||
        (ctx->error == X509_V_ERR_INVALID_CA) ||
        (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED) ||
        (ctx->error == X509_V_ERR_INVALID_PURPOSE) ||
        (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ||
        (ctx->error == X509_V_ERR_CRL_HAS_EXPIRED) ||
        (ctx->error == X509_V_ERR_CRL_NOT_YET_VALID) ||
        (ctx->error == X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION)) {
      ok = 1;
    }
  }

  ERR_clear_error();
  return ok;
}


static X509 *load_cert(BIO *err, int type, const char *filename)
{
  X509 *x509 = NULL;
  PKCS12 *p12 = NULL;
  BIO *cert;

  if ((cert = BIO_new_file(filename, "r")) == NULL) {
    BIO_printf(err, "Error opening certificate file %s\n", filename);
    ERR_print_errors(err);
    goto _exit;
  }

  switch (type) {

  case FILETYPE_ASN1:
    x509 = d2i_X509_bio(cert, NULL);
    break;

  case FILETYPE_PEM:
    x509 = PEM_read_bio_X509(cert, NULL, NULL, NULL);
    break;
    
  case FILETYPE_PKCS12:
    p12 = d2i_PKCS12_bio(cert, NULL);
    PKCS12_parse(p12, NULL, NULL, &x509, NULL);
    PKCS12_free(p12);
    break;

  default:
    goto _exit;
  }

_exit:
  if (cert != NULL)
    BIO_free(cert);

  if (x509 == NULL) {
    BIO_printf(err, "Fail to load certificate\n");
    ERR_print_errors(err);
  }

  return x509;
}


int auth_cert_verify_get_purpose_id(const char *purpose_name)
{
  X509_PURPOSE *xpurpose = NULL;
  int i;
  
  if ((i = X509_PURPOSE_get_by_sname((char *) purpose_name)) < 0) {
    BIO_printf(bio_err, "unrecognized purpose\n");
    return -1;
  }
  xpurpose = X509_PURPOSE_get0(i);
  return X509_PURPOSE_get_id(xpurpose);
}


X509_STORE *auth_cert_verify_init(const char *ca_path, const char *ca_file)
{
  X509_STORE *ret = NULL, *cert_ctx = NULL;
  X509_LOOKUP *lookup = NULL;
  int i;

  if ((cert_ctx = X509_STORE_new()) == NULL)
    goto _exit;

  X509_STORE_set_verify_cb_func(cert_ctx, verify_cb);

  /* lookup CA file */
  lookup = X509_STORE_add_lookup(cert_ctx, X509_LOOKUP_file());
  if (lookup == NULL) {
    BIO_printf(bio_err, "Error adding X509 store lookup: X509_LOOKUP_file\n");
    goto _exit;
  }
  if (ca_file) {
    i = X509_LOOKUP_load_file(lookup, ca_file, X509_FILETYPE_PEM);
    if (!i) {
      BIO_printf(bio_err, "Error loading file %s\n", ca_file);
      ERR_print_errors(bio_err);
      goto _exit;
    }
  } else X509_LOOKUP_load_file(lookup, NULL, X509_FILETYPE_DEFAULT);

  /* lookup CA path */
  lookup = X509_STORE_add_lookup(cert_ctx, X509_LOOKUP_hash_dir());
  if (lookup == NULL) {
    BIO_printf(bio_err, "Error adding X509 store lookup: X509_LOOKUP_lookup\n");
    goto _exit;
  }
  if (ca_path) {
    i = X509_LOOKUP_add_dir(lookup, ca_path, X509_FILETYPE_PEM);
    if (!i) {
      BIO_printf(bio_err, "Error loading directory %s\n", ca_path);
      ERR_print_errors(bio_err);
      goto _exit;
    }
  } else X509_LOOKUP_add_dir(lookup, NULL, X509_FILETYPE_DEFAULT);

  ret = cert_ctx;

_exit:
  return ret;
}


int auth_verify_cert(X509_STORE *ctx, const char *filename, int purpose,
                     int vflags)
{
  X509_STORE_CTX *csc;
  X509 *x509 = NULL;
  int i = 0, ret = 0;

  if ((x509 = load_cert(bio_err, FILETYPE_PEM, filename)) == NULL)
    goto _exit;

  if ((csc = X509_STORE_CTX_new()) == NULL) {
    ERR_print_errors(bio_err);
    goto _exit;
  }
  
  X509_STORE_set_flags(ctx, vflags);
  if (X509_STORE_CTX_init(csc, ctx, x509, NULL) == 0) {
    ERR_print_errors(bio_err);
    goto _exit;
  }

  if (purpose >= 0) {
    X509_STORE_CTX_set_purpose(csc, purpose);
  }

  i = X509_verify_cert(csc);
  X509_STORE_CTX_free(csc);
  
  ret = 0;

_exit:
  if (i>0) {
    printf("Certificate verification OK\n");
    ret = 1;
  }  else
    ERR_print_errors(bio_err);
  if (x509 != NULL) X509_free(x509);

  return ret;
}
