#include <linux/module.h>
#include <linux/skbuff.h>
#include <net/checksum.h>
#include <linux/random.h>
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_EDGEMARK.h>

#define EDGEMARK_STATE_STOPPED  0
#define EDGEMARK_STATE_STARTED  1

struct edgemark_state {
  int state;
  unsigned long long total_cnt;         // total processed packet count
  unsigned long long skipopt_cnt;       // skipped packet count
  unsigned long long mark_cnt;          // marked packet count
  unsigned long long overwritten_cnt;   // mark overwritten count
  spinlock_t lock;
};

unsigned long markprob_unscaled;

static struct edgemark_state em_state = {
  state           : EDGEMARK_STATE_STOPPED,
  total_cnt       : 0,
  skipopt_cnt     : 0,
  mark_cnt        : 0,
  overwritten_cnt : 0,
  lock            : SPIN_LOCK_UNLOCKED
};

#define edgemark_lock_state() spin_lock_bh(&(em_state.lock))
#define edgemark_unlock_state() spin_unlock_bh(&(em_state.lock))
#define edgemark_start() em_state.state = EDGEMARK_STATE_STARTED
#define edgemark_stop() em_state.state = EDGEMARK_STATE_STOPPED
#define edgemark_state() em_state.state

inline void
edgemark_reset()
{
  em_state.total_cnt = 0;
  em_state.skipopt_cnt = 0;
  em_state.mark_cnt = 0;
  em_state.overwritten_cnt = 0;
}

/*
 *  Check the edgemark message type of a packet header
 *
 *  returns:
 *    EDGEMARK_MSGTYPE_START : start command
 *    EDGEMARK_MSGTYPE_STOP  : stop command
 *    EDGEMARK_MSGTYPE_RESET : reset command
 *    EDGEMARK_MSGTYPE_MARK  : makred packet   
 *    0                      : packet without IP options
 *    -1                     : packet with non-mark IP options
 */
inline __u8
edgemark_msgtype(struct iphdr *iph)
{
  struct edgemark_opt *opt = EDGEMARK_OPTPTR(iph);
  if ((iph->ihl == (EDGEMARK_IPHLEN >> 2)) &&
      (opt->type == IPOPT_EDGEMARK) &&
      (opt->length == EDGEMARK_OPTLEN))
  {
    return opt->msgtype;
  }

  if (iph->ihl == (sizeof(struct iphdr) >> 2))
    return 0;

  return -1;
}

/*
 *  Get the IP address of a net device
 */
inline static unsigned long
get_ip_address(const struct net_device *dev)
{
  struct in_ifaddr *ifaddr;
	struct in_device *in_dev;
  unsigned long ret = 0;
  
//  rtnl_lock();
  
	if ((in_dev = __in_dev_get(dev)) == NULL) {
    goto _exit;
	}

	if ((ifaddr = in_dev->ifa_list) == NULL) {
    goto _exit;
	}	

  ret = ifaddr->ifa_local;

_exit:
//  rtnl_unlock();

  return ret;
}	

/*
 *  Edge-marking routine
 */
inline static int
proc_edgemark(struct sk_buff **pskb,
              unsigned long addr,
              int domark)
{
  struct sk_buff *newskb, *skb;
  struct iphdr *iph;
  int iphlen;
  struct edgemark_opt *opt;
  
  skb = *pskb;
  iph = skb->nh.iph;
  iphlen = iph->ihl << 2;

  em_state.total_cnt++;
          
  switch (iphlen) {
    
  case sizeof(struct iphdr):  // unmarked packet (normal IP header)
    if (domark) {
      em_state.mark_cnt++;
      
      newskb = skb_copy_expand(skb,
                               skb_headroom(skb),
                               EDGEMARK_OPTLEN,
                               GFP_ATOMIC);  

      if (newskb == NULL) {
        printk(KERN_WARNING "ip_queue: OOM "
               "in mangle, dropping packet\n");
        return -ENOMEM;
      }

      skb_put(newskb, EDGEMARK_OPTLEN);
      
      memset(newskb->data + sizeof(struct iphdr), 0, EDGEMARK_OPTLEN);
      memcpy(newskb->data + sizeof(struct iphdr) + EDGEMARK_OPTLEN,
             skb->data + sizeof(struct iphdr),
             skb->len - sizeof(struct iphdr));
  
      iph = newskb->nh.iph;
      iph->ihl = (EDGEMARK_IPHLEN >> 2);
      iph->tot_len = htons(ntohs(iph->tot_len) + EDGEMARK_OPTLEN);

      opt = (struct edgemark_opt *) (newskb->data + sizeof(struct iphdr));

      opt->type = IPOPT_EDGEMARK;
      opt->length = EDGEMARK_OPTLEN;
      opt->msgtype = EDGEMARK_MSGTYPE_MARK;

      opt->mark_start = addr;
      opt->mark_end = 0; // htonl(0)
      opt->mark_distance = 0;

      // recalculate IP header checksum
      iph->check = 0;
      iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);

      // mark sk_buff as altered
      newskb->nfcache |= NFC_ALTERED;

      if (skb->sk)
        skb_set_owner_w(newskb, skb->sk);
  
      kfree_skb(skb);
  
      *pskb = newskb;
      DEBUGP("add_edgemark: NEW (s:%u.%u.%u.%u, e:%u.%u.%u.%u, d:%d)\n",
             NIPQUAD(opt->mark_start),
             NIPQUAD(opt->mark_end),
             opt->mark_distance);
    }
    break;
    
  case EDGEMARK_IPHLEN:  // marked packet (IP header with IP options)
    opt = (struct edgemark_opt *) (skb->data + sizeof(struct iphdr));
    
    if (domark) {
      DEBUGP("add_edgemark: RST (s:%u.%u.%u.%u, e:%u.%u.%u.%u, d:%d) ->",
             NIPQUAD(opt->mark_start),
             NIPQUAD(opt->mark_end),
             opt->mark_distance);

      em_state.overwritten_cnt++;
      opt->mark_start = addr;
      opt->mark_end = 0; // htonl(0)
      opt->mark_distance = 0;
      
      DEBUGP("(s:%u.%u.%u.%u, e:%u.%u.%u.%u, d:%d)\n",
             NIPQUAD(opt->mark_start),
             NIPQUAD(opt->mark_end),
             opt->mark_distance);
      
    } else {
      
      DEBUGP("add_edgemark: ADD (s:%u.%u.%u.%u, e:%u.%u.%u.%u, d:%d) ->",
             NIPQUAD(opt->mark_start),
             NIPQUAD(opt->mark_end),
             opt->mark_distance);
      
      if (opt->mark_distance == 0) {
        opt->mark_end = addr;
      }
      opt->mark_distance++;
      
      DEBUGP("(s:%u.%u.%u.%u, e:%u.%u.%u.%u, d:%d)\n",
             NIPQUAD(opt->mark_start),
             NIPQUAD(opt->mark_end),
             opt->mark_distance);
    }
    
    iph->check = 0;
    iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);
    
    // mark sk_buff as altered      
    skb->nfcache |= NFC_ALTERED;
    break;
  }

  return 0;
}

static unsigned int
target(struct sk_buff **pskb,
       unsigned int hooknum,
       const struct net_device *in,
       const struct net_device *out,
       const void *targinfo,
       void *userinfo)
{
	const struct ipt_edgemark_target_info *edgemarkinfo = targinfo;
  unsigned long addr;
  u_int8_t rnum[1024];
  int ridx = 0;

  switch (edgemark_msgtype((*pskb)->nh.iph)) {
  case EDGEMARK_MSGTYPE_START:
    edgemark_lock_state();
    edgemark_start();
    edgemark_unlock_state();
    break;
    
  case EDGEMARK_MSGTYPE_STOP:
    edgemark_lock_state();
    edgemark_stop();
    edgemark_unlock_state();
    break;

  case EDGEMARK_MSGTYPE_RESET:
    edgemark_lock_state();
    edgemark_reset();
    edgemark_unlock_state();
    break;

  case -1:  // packet with IP options
    edgemark_lock_state();
    // existing options are kept
    em_state.skipopt_cnt++;
    edgemark_unlock_state();
    break;
    
  default:  // unmarked or marked packet
    edgemark_lock_state();
    if (edgemark_state() == EDGEMARK_STATE_STOPPED) {
      edgemark_unlock_state();
      break;
    }
    edgemark_unlock_state();
    
    if (out != NULL) {
      if ((addr = get_ip_address(out)) != 0) {
        // get random numbers from the kernel random number generation routine
        if (ridx == 0)
          get_random_bytes((void *)(&rnum), sizeof(rnum));
        
        proc_edgemark(pskb, addr, (rnum[ridx] <= edgemarkinfo->markprob));
        
        if ((++ridx) == sizeof(rnum))
          ridx = 0;
      }
    }
  }

  return IPT_CONTINUE;
}

static int
checkentry(const char *tablename,
           const struct ipt_entry *e,
           void *targinfo,
           unsigned int targinfosize,
           unsigned int hook_mask)
{
	const struct ipt_edgemark_target_info *edgemarkinfo = targinfo;

  if (targinfosize != IPT_ALIGN(sizeof(struct ipt_edgemark_target_info))) {
		printk(KERN_WARNING "EDGEMARK: targinfosize %u != %Zu\n",
		       targinfosize,
		       IPT_ALIGN(sizeof(struct ipt_edgemark_target_info)));
		return 0;
	}

	/* must be  1 <= average % <= 99 */
	/* 1  x 2.55 = 2   */
	/* 99 x 2.55 = 252 */
	if ((edgemarkinfo->markprob < 2) || (edgemarkinfo->markprob > 252)) {
		printk("ipt_EDGEMARK:  invalid marking probaility %lu\n",
           edgemarkinfo->markprob);
		return 0;
	}

	markprob_unscaled = edgemarkinfo->markprob_unscaled;

  if (strcmp(tablename, "mangle") == 0) {
    if (hook_mask & ~((1 << NF_IP_POST_ROUTING)
                      | (1 << NF_IP_FORWARD)
                      | (1 << NF_IP_LOCAL_OUT)))
    {
      printk(KERN_WARNING "EDGEMARK: bad hook in \"mangle\" table\n");
      return 0;
    }
  } else if (strcmp(tablename, "filter") == 0) {
    if (hook_mask & ~((1 << NF_IP_PRE_ROUTING)
                      | (1 << NF_IP_LOCAL_IN)))
    {
      printk(KERN_WARNING "EDGEMARK: bad hook in \"filter\" table\n");
      return 0;
    }
  } else {
		printk(KERN_WARNING "EDGEMARK: can only be called from \"filter\" or \"mangle\" table, not \"%s\"\n",
           tablename);    
  }
  
	return 1;
}

static int
edgemark_proc_read(char *page,
                   char **start,
                   off_t off,
                   int count,
                   int *eof,
                   void *data)
{
  int len = 0;
  
  MOD_INC_USE_COUNT;

  edgemark_lock_state();

  len += sprintf(page+len,
                 "marking probability    : %lu%%\n"
                 "state                  : %s\n\n"
                 "total normal packets   : %llu\n"
                 "marked packets         : %llu\n"
                 "unmarked packets       : %llu\n"
                 "mark overrwritten      : %llu\n"
                 "skipped (with options) : %llu\n",
                 markprob_unscaled,
                 (edgemark_state() == EDGEMARK_STATE_STOPPED) ? "STOPPED" : "STARTED",
                 em_state.total_cnt,
                 em_state.mark_cnt,
                 em_state.total_cnt - em_state.mark_cnt,
                 em_state.overwritten_cnt,
                 em_state.skipopt_cnt);

  edgemark_unlock_state();

  MOD_DEC_USE_COUNT;

  return len;
}

static int
edgemark_proc_write(struct file *file,
                    const char *buffer,
                    unsigned long count,
                    void *data)
{
  MOD_INC_USE_COUNT;

  if (count == 6) {
    if (strncmp(buffer, "start", 5) == 0) {
      printk("EDGEMARK: start\n");
      edgemark_lock_state();
      edgemark_start();
      edgemark_unlock_state();
    } else if (strncmp(buffer, "reset", 5) == 0) {
      printk("EDGEMARK: reset\n");
      edgemark_lock_state();
      edgemark_reset();
      edgemark_unlock_state();
    }
  } else if (count == 5) {
    if (strncmp(buffer, "stop", 4) == 0) {
      printk("EDGEMARK: stop\n");
      edgemark_lock_state();
      edgemark_stop();
      edgemark_unlock_state();
    }
  }
  
  MOD_DEC_USE_COUNT;

  return count;
}

static struct ipt_target ipt_edgemark_reg
= { { NULL, NULL }, "EDGEMARK", target, checkentry, NULL, THIS_MODULE };

static const char EDGEMARK_PROCFS_PATH[] = "edgemark";
extern struct proc_dir_entry *proc_net;

static int __init
init(void)
{
  struct proc_dir_entry *proc;

  proc = create_proc_entry(EDGEMARK_PROCFS_PATH,
                           S_IRUSR | S_IRGRP | S_IROTH, proc_net);
  if (!proc) {
    return -ENOMEM;
  }
  proc->read_proc = edgemark_proc_read;
  proc->write_proc = edgemark_proc_write;
  proc->owner = THIS_MODULE;

  if (ipt_register_target(&ipt_edgemark_reg))
		return -EINVAL;

  em_state.lock = SPIN_LOCK_UNLOCKED;
  edgemark_stop();
  
	return 0;
}

static void __exit
fini(void)
{
	ipt_unregister_target(&ipt_edgemark_reg);
  remove_proc_entry(EDGEMARK_PROCFS_PATH, proc_net);  
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
