#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_SNAPSHOT.h>

static struct snapshot_rec ssrec;

static unsigned int
ipt_ssvictim_target(struct sk_buff **pskb,
                    unsigned int hooknum,
                    const struct net_device *in,
                    const struct net_device *out,
                    const void *targinfo,
                    void *userinfo)
{
  struct iphdr *iph = (*pskb)->nh.iph;
  struct snapshot_opt *opt = SNAPSHOT_OPTPTR(iph);  
  int iphlen = iph->ihl << 2;
  __u8 *data = ((__u8 *) iph) + iphlen;
  int i, flags;

  ssrec_lock(ssrec, flags);
  
  switch (snapshot_msgtype(iph)) {
  case SNAPSHOT_MSGTYPE_RESET:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSVICTIM", "RESET", in, iph, opt);
      DEBUGP("SSVICTIM: reset\n");
      snapshot_reset(&ssrec);
    }
    break;

  case SNAPSHOT_MSGTYPE_CLEAR:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSVICTIM", "CLEAR", in, iph, opt);
      if (ssrec.local.state != SNAPSHOT_STATE_UNINITIALIZED) {
        DEBUGP("SSVICTIM: clear\n");
        snapshot_clear(&ssrec);
      }
    }
    break;
    
  case SNAPSHOT_MSGTYPE_VMAP:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSVICTIM", "VMAP", in, iph, opt);
      DEBUGP("SSVICTIM: VMAP size = %d\n", (*pskb)->len - iphlen);
      
      if ((ssrec.local.state == SNAPSHOT_STATE_UNINITIALIZED) ||
          (ssrec.local.state == SNAPSHOT_STATE_FINISHED))
      {
        ssrec.local.state = SNAPSHOT_STATE_READY;
        ssrec.local.addr = iph->daddr; // trick: get local addr here
        snapshot_parse_vmap(&ssrec, data, (*pskb)->len - iphlen);
      }
      
      snapshot_send_ack(&ssrec, iph->daddr, iph->saddr,
                        SNAPSHOT_MSGTYPE_VMAP_ACK);
    }
    break;

  case SNAPSHOT_MSGTYPE_SMAP:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSVICTIM", "SMAP", in, iph, opt);
      DEBUGP("SSVICTIM: SMAP size = %d\n", (*pskb)->len - iphlen);
      
      if ((ssrec.local.state == SNAPSHOT_STATE_UNINITIALIZED) ||
          (ssrec.local.state == SNAPSHOT_STATE_FINISHED))
      {
        snapshot_parse_smap(&ssrec, data, (*pskb)->len - iphlen);
      }
      
      snapshot_send_ack(&ssrec, iph->daddr, iph->saddr,
                        SNAPSHOT_MSGTYPE_SMAP_ACK);
    }
    break;
    
  case SNAPSHOT_MSGTYPE_CONTROL:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSVICTIM", "CONTROL", in, iph, opt);
      
      if ((ssrec.local.state == SNAPSHOT_STATE_READY) ||
          (ssrec.local.state == SNAPSHOT_STATE_FINISHED))
      {
        ssrec.local.state = SNAPSHOT_STATE_STARTED;
        ssrec.local.cnt = ssrec.packetcnt;
        ssrec.seqno = ntohl(opt->seqno);
        ssrec.finished = 0;

        if (ssrec.statcnt != 0)
          snapshot_send_stat(&ssrec, ssrec.local.addr, ssrec.local.addr);
        
        for_each_link(&ssrec, i) {
          ssrec.link[i].state = SNAPSHOT_STATE_STARTED;
          ssrec.link[i].cnt = 0;
          snapshot_send_marker(&ssrec, ssrec.local.addr, ssrec.link[i].addr);
        }

        for_each_stat(&ssrec, i) {
          ssrec.stat[i].cnt = 0;
        }
      }

      snapshot_send_ack(&ssrec, ssrec.local.addr, iph->saddr,
                        SNAPSHOT_MSGTYPE_CONTROL_ACK);
    }
    break;
        
  case SNAPSHOT_MSGTYPE_MARKER:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSVICTIM", "MARKER", in, iph, opt);
    
      i = snapshot_find_link_by_dev_addr(&ssrec, (unsigned char *) in->dev_addr);
      if (i != -1) {
        if (ssrec.link[i].state == SNAPSHOT_STATE_STARTED) {
          ssrec.link[i].state = SNAPSHOT_STATE_FINISHED;
        }
      } else {
        DEBUGP("SSVICTIM: no match link\n");
      }
    }
    break;

  case SNAPSHOT_MSGTYPE_SNAPSHOT:
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {
      DEBUGP_SSMSG("SSVICTIM", "SNAPSHOT", in, iph, opt);

      if (ssrec.local.state == SNAPSHOT_STATE_STARTED) {
        DEBUGP("SSVICTIM: snapshot count = %d\n", ssrec.finished);
        ssrec.finished++;
        if (ssrec.finished == ssrec.snapshotcnt) {
          DEBUGP("SSVICTIM: send SNAPSHOT (%u)\n", ssrec.seqno);
          ssrec.local.state = SNAPSHOT_STATE_FINISHED;
          snapshot_send_snapshot(&ssrec, ssrec.local.addr, ssrec.local.addr);
        }
      }
    }
    break;
    
  default:
    // only INPUT packets here
    if ((hooknum == NF_IP_LOCAL_IN) && (in != NULL)) {

      // all snapshot messages, not matter valid or not, are not counted
      if (iph->protocol != IPPROTO_SNAPSHOT) {

        ssrec.packetcnt++;

        DEBUGP("SSROUTER: recv PACKET from %u.%u.%u.%u/", NIPQUAD(iph->saddr));
        DEBUGP_DEV_ADDR((unsigned char *) in->dev_addr, in->addr_len);
      
        i = snapshot_find_link_by_dev_addr(&ssrec, (unsigned char *) in->dev_addr);
        if (i != -1) {
          if (ssrec.link[i].state == SNAPSHOT_STATE_STARTED) {
            ssrec.link[i].cnt++;
          }
        } else {
          DEBUGP("SSVICTIM: no match link\n");
        }
 
        i = snapshot_find_stat_by_addr(&ssrec, iph->saddr);
        if (i != -1) {
          ssrec.stat[i].cnt++;
        }
      }
    }
  }

  ssrec_unlock(ssrec, flags);

  return IPT_CONTINUE;
}

static int
ipt_ssvictim_checkentry(const char *tablename,
                        const struct ipt_entry *e,
                        void *targinfo,
                        unsigned int targinfosize,
                        unsigned int hook_mask)
{

  if (hook_mask & ~(1 << NF_IP_LOCAL_IN)) {
		DEBUGP("SSVICTIM: bad hook\n");
		return 0;
	}

	if (targinfosize != IPT_ALIGN(0)) {
		DEBUGP("SSVICTIM: targinfosize %u != 0\n", targinfosize);
		return 0;
	}

	return 1;
}

static int
ssvictim_proc_read(char *page,
                   char **start,
                   off_t off,
                   int count,
                   int *eof,
                   void *data)
{
  int len, flags;
  
  MOD_INC_USE_COUNT;

  ssrec_lock(ssrec, flags);
  len = snapshot_print_rec(page, &ssrec);
  ssrec_unlock(ssrec, flags);
  
  MOD_DEC_USE_COUNT;

  return len;
}

static struct ipt_target ipt_ssvictim_reg
= { { NULL, NULL }, "SSVICTIM", ipt_ssvictim_target, ipt_ssvictim_checkentry,
    NULL, THIS_MODULE };

static const char SSVICTIM_PROCFS_PATH[] = "snapshot_victim";
extern struct proc_dir_entry *proc_net;

static int __init
init(void)
{
  struct proc_dir_entry *proc;
  int ret;
  
  ret = snapshot_rec_alloc_internal(&ssrec, MAX_LINK_LIMIT, MAX_STAT_LIMIT);
  if (ret < 0) {
    return ret;
  }
  
  proc = create_proc_entry(SSVICTIM_PROCFS_PATH, S_IRUSR | S_IRGRP | S_IROTH, proc_net);
  if (!proc) {
    snapshot_rec_free_internal(&ssrec);
    return -ENOMEM;
  }
  proc->read_proc = ssvictim_proc_read;
  proc->write_proc = NULL;
  proc->owner = THIS_MODULE;
  
	return ipt_register_target(&ipt_ssvictim_reg);
}

static void __exit
fini(void)
{
	ipt_unregister_target(&ipt_ssvictim_reg);
  remove_proc_entry(SSVICTIM_PROCFS_PATH, proc_net);
  snapshot_rec_free_internal(&ssrec);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
