#include <linux/skbuff.h>
#include <net/checksum.h>
#include <linux/pkt_sched.h>
#include <linux/vmalloc.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_SNAPSHOT.h>

/*
  !!! ATTENTION !!!

  all functions here DO NOT lock snapshot_rec,
  locking is left to outer event loop
*/

inline __u8 snapshot_msgtype(struct iphdr *iph)
{
  struct snapshot_opt *opt = SNAPSHOT_OPTPTR(iph);
  if ((iph->protocol == IPPROTO_SNAPSHOT) &&
      (iph->ihl == (SNAPSHOT_IPHLEN >> 2)) &&
      (opt->type == IPOPT_SNAPSHOT) &&
      (opt->length == SNAPSHOT_OPTLEN))
  {
    return opt->msgtype;
  }
  return 0;
}

// stolen from ip_output.c
static inline int ip_finish_output2(struct sk_buff *skb)
{
  struct dst_entry *dst = skb->dst;
  struct hh_cache *hh = dst->hh;
 
  if (hh) {
    read_lock_bh(&hh->hh_lock);
    memcpy(skb->data - 16, hh->hh_data, 16);
    read_unlock_bh(&hh->hh_lock);
    skb_push(skb, hh->hh_len);
    return hh->hh_output(skb);
  } else if (dst->neighbour)
    return dst->neighbour->output(skb);
 
  if (net_ratelimit())
    printk(KERN_DEBUG "ip_finish_output2: No header cache and no neighbour!\n");
  kfree_skb(skb);
  return -EINVAL;
}

static inline struct sk_buff*
snapshot_generate_skb(__u32 saddr, __u32 daddr, __u8 type, int datalen,
                      __u32 seqno)
{
  struct sk_buff *skb = NULL;
  struct iphdr *iph;
  struct rtable *rt;
  struct snapshot_opt *opt;
  int hh_len, ret;
  int totlen;
  
  // find route
  if ((ret = ip_route_output(&rt, daddr, saddr, 0, 0)))
  {
    DEBUGP("snapshot_generate_skb: route failed: %d\n", ret);
    goto out;
	}

  if (rt->u.dst.dev == NULL) {
    DEBUGP("snapshot_generate_skb: route failed: %d\n", ret);
    goto out;
  }
  
  // length
  totlen = SNAPSHOT_IPHLEN + datalen;
  hh_len = (rt->u.dst.dev->hard_header_len + 15)&~15;
  
  // hh_len for headspace
  skb = alloc_skb(totlen + hh_len, GFP_ATOMIC);
  if (skb == NULL) {
    goto out;
  }

  skb_reserve(skb, hh_len);

  // assign device
  skb->dev = rt->u.dst.dev;
  skb->dst = dst_clone(&rt->u.dst);  // <- ?????

  skb->protocol = __constant_htons(ETH_P_IP);
//  skb->priority = TC_PRIO_MAX;
  skb->priority = TC_PRIO_BESTEFFORT;
  

  // build ethernet header
  /*
  if (skb->dev->hard_header &&
	    skb->dev->hard_header(skb, skb->dev, ETH_P_IP, skb->dev->dev_addr,
                       skb->dev->dev_addr, skb->len) < 0)
		goto out_free;

	if (skb->dev->rebuild_header && skb->dev->rebuild_header(skb) < 0)
		goto out_free;
  */
  skb->nh.iph = (struct iphdr *) skb_put(skb, totlen);

  // mark skb as altered
  skb->nfcache |= NFC_ALTERED;

  // default IP header
  iph = skb->nh.iph;
  iph->version = IPVERSION;
  iph->ihl = SNAPSHOT_IPHLEN >> 2;
  iph->tos = 0;
  iph->tot_len = htons(totlen);
  iph->id = 0;
  iph->frag_off = 0;
  iph->ttl = MAXTTL;
  iph->saddr = saddr;
  iph->daddr = daddr;
  iph->protocol = IPPROTO_SNAPSHOT;

  // IP options
  opt = SNAPSHOT_OPTPTR(iph);
  opt->type = IPOPT_SNAPSHOT;
  opt->length = SNAPSHOT_OPTLEN;
  opt->msgtype = type;
  opt->padding = 0;
  opt->seqno = htonl(seqno);

/*  
  goto out;

out_free:
  kfree_skb(skb);
  skb = NULL;
*/
  
out:
  return skb;
}

int snapshot_send_marker(struct snapshot_rec *rec, __u32 saddr, __u32 daddr)
{
  struct sk_buff *skb;
  struct iphdr *iph;


  // generate default snapshot skb
  skb = snapshot_generate_skb(saddr, daddr, SNAPSHOT_MSGTYPE_MARKER,
                              0, rec->seqno);
  if (skb == NULL)
    return -1;

  DEBUGP("snapshot_send_marker: to %u.%u.%u.%u\n", NIPQUAD(daddr));

  iph = skb->nh.iph;
  
  // calculate IP header checksum
  ip_send_check(iph);

  return NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, skb->dev,
                 ip_finish_output2);

/*  
  // transmit buffer via device
  if (dev_queue_xmit(skb) < 0) {
    DEBUGP("snapshot_send_marker: dev_queue_xmit failed\n");
    return -1;
  }

  return 0;
*/
}

int snapshot_send_snapshot(struct snapshot_rec *rec, __u32 saddr, __u32 daddr)
{
  struct sk_buff *skb;
  struct iphdr *iph;
  struct snapshot_msghdr_snapshot *hdr;
  __u32 *data;
  int datalen;
  int i, j;

  datalen = sizeof(__u32) + (rec->linkcnt+1) * 2 * sizeof(__u32);

  // generate default snapshot skb
  skb = snapshot_generate_skb(saddr, daddr, SNAPSHOT_MSGTYPE_SNAPSHOT,
                              datalen, rec->seqno);
  if (skb == NULL)
    return -1;

  iph = skb->nh.iph;
  
  // IP payload
  hdr = (struct snapshot_msghdr_snapshot *) SNAPSHOT_DATAPTR(iph);
  hdr->entrycnt = htons(rec->linkcnt+1);
  hdr->padding = 0;
  
  data = (__u32 *) (SNAPSHOT_DATAPTR(iph) + 4);
  data[0] = rec->local.addr;
  data[1] = htonl(rec->local.cnt);
  j = 2;
  for (i=0; i<rec->linkcnt; i++) {
    data[j] = rec->link[i].addr;
    data[j+1] = htonl(rec->link[i].cnt);
    j += 2;
  }
  
  // calculate IP header checksum
  ip_send_check(iph);

  return NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, skb->dev,
                 ip_finish_output2);

/*  
  // transmit buffer via device
  if (dev_queue_xmit(skb) < 0)
    return -1;

  return 0;
*/
}

int snapshot_send_stat(struct snapshot_rec *rec, __u32 saddr, __u32 daddr)
{
  struct sk_buff *skb;
  struct iphdr *iph;
  struct snapshot_msghdr_stat *hdr;
  __u32 *data;
  int datalen;
  int i, j;

  datalen = sizeof(__u32) + (rec->statcnt) * 2 * sizeof(__u32);

  // generate default snapshot skb
  skb = snapshot_generate_skb(saddr, daddr, SNAPSHOT_MSGTYPE_STAT,
                              datalen, rec->seqno);
  if (skb == NULL)
    return -1;

  iph = skb->nh.iph;
  
  // IP payload
  hdr = (struct snapshot_msghdr_stat *) SNAPSHOT_DATAPTR(iph);
  hdr->entrycnt = htons(rec->statcnt);
  hdr->padding = 0;
  
  data = (__u32 *) (SNAPSHOT_DATAPTR(iph) + 4);
  j = 0;
  for (i=0; i<rec->statcnt; i++) {
    data[j] = rec->stat[i].addr;
    data[j+1] = htonl(rec->stat[i].cnt);
    j += 2;
  }
  
  // calculate IP header checksum
  ip_send_check(iph);


  return NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, skb->dev,
                 ip_finish_output2);

/*  
  // transmit buffer via device
  if (dev_queue_xmit(skb) < 0)
    return -1;

  return 0;
*/
}

int snapshot_send_ack(struct snapshot_rec *rec, __u32 saddr, __u32 daddr,
                      int msgtype)
{
  struct sk_buff *skb;
  struct iphdr *iph;


  // generate default snapshot skb
  skb = snapshot_generate_skb(saddr, daddr, msgtype, 0, rec->seqno);
  if (skb == NULL)
    return -1;

  DEBUGP("snapshot_send_ack: to %u.%u.%u.%u\n", NIPQUAD(daddr));

  iph = skb->nh.iph;
  
  // calculate IP header checksum
  ip_send_check(iph);


  return NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, skb->dev,
                 ip_finish_output2);


/*  
  // transmit buffer via device
  if (dev_queue_xmit(skb) < 0) {
    DEBUGP("snapshot_send_marker: dev_queue_xmit failed\n");
    return -1;
  }
  
  return 0;
*/
}

int snapshot_parse_vmap(struct snapshot_rec *rec, __u8 *data, int len)
{
  struct rtable *rt;
  struct net_device *dev;
  struct snapshot_msghdr_vmap *hdr = (struct snapshot_msghdr_vmap *) data;
  __u32 *map = (__u32 *) (data + sizeof(struct snapshot_msghdr_vmap));
  int i, mapcnt = 0;
  
  // checking on entry count is skipped !!!
  rec->snapshotcnt = ntohs(hdr->snapshotcnt);
  
  // exclude message header
  len = (len - sizeof(struct snapshot_msghdr_vmap)) >> 2;

  for (i=0; (i<len) && (i<rec->maxlinkcnt); i++) {
    if (map[i] == 0)
      break;

    // trick: saddr = 0
    if (ip_route_output(&rt, map[i], 0, 0, 0)) {
      DEBUGP("snapshot_parse_vmap: map[%d]: %u.%u.%u.%u: route failed\n",
             i, NIPQUAD(map[i]));
      continue;
    }

    DEBUGP("snapshot_parse_vmap: map[%d]: %u.%u.%u.%u: route ok\n",
           i, NIPQUAD(map[i]));
    dev = rt->u.dst.dev;
    rec->link[i].dev_addrlen = dev->addr_len;
    memcpy(rec->link[i].dev_addr, dev->dev_addr, dev->addr_len);
    rec->link[i].addr = map[i];
    rec->link[i].state = SNAPSHOT_STATE_READY;
    mapcnt++;
  }

  rec->linkcnt = mapcnt;
  return mapcnt;
}

int snapshot_parse_rmap(struct snapshot_rec *rec, __u8 *data, int len)
{
  struct rtable *rt;
  struct net_device *dev;
  struct snapshot_msghdr_rmap *hdr = (struct snapshot_msghdr_rmap *) data;
  __u32 *map;
  int i, mapcnt = 0;

  rec->victim_addr = *((__u32 *) (data + sizeof(struct snapshot_msghdr_rmap)));
  
  // checking on entry count is skipped !!!
  
  // hdr->connected (1: direct-connected to victim, 0: otherwise)
  if (hdr->connected == 0) {
    map = (__u32 *) (data + sizeof(struct snapshot_msghdr_rmap) + sizeof(__u32));
    // exclude message header and victim addr
    len = (len - sizeof(struct snapshot_msghdr_rmap) - sizeof(__u32)) >> 2;
  } else {
    map = (__u32 *) (data + sizeof(struct snapshot_msghdr_rmap));
    // exclude message header
    len = (len - sizeof(struct snapshot_msghdr_rmap)) >> 2;
  }

  for (i=0; (i<len) && (i<rec->maxlinkcnt); i++) {
    if (map[i] == 0)
      break;

    // trick: saddr = 0
    if (ip_route_output(&rt, map[i], 0, 0, 0)) {
      DEBUGP("snapshot_parse_rmap: map[%d]: %u.%u.%u.%u: route failed ... try to resolve by victim address\n",
             i, NIPQUAD(map[i]));
      if (ip_route_output(&rt, rec->victim_addr, 0, 0, 0)) {
        DEBUGP("snapshot_parse_rmap: map[%d]: %u.%u.%u.%u: route failed\n",
               i, NIPQUAD(map[i]));
        continue;
      }
    }

    DEBUGP("snapshot_parse_rmap: map[%d]: %u.%u.%u.%u: router ok\n",
           i, NIPQUAD(map[i]));
    dev = rt->u.dst.dev;
    rec->link[i].dev_addrlen = dev->addr_len;
    memcpy(rec->link[i].dev_addr, dev->dev_addr, dev->addr_len);
    rec->link[i].addr = map[i];
    rec->link[i].state = SNAPSHOT_STATE_READY;
    mapcnt++;
  }

  rec->linkcnt = mapcnt;
  return mapcnt;
}

int snapshot_parse_smap(struct snapshot_rec *rec, __u8 *data, int len)
{
//  struct snapshot_msghdr_smap *hdr = (struct snapshot_msghdr_smap *) data;
  __u32 *map = (__u32 *) (data + sizeof(struct snapshot_msghdr_smap));
  int i, mapcnt = 0;
  
  // checking on entry count is skipped !!!
  
  // exclude message header
  len = (len - sizeof(struct snapshot_msghdr_smap)) >> 2;

  for (i=0; (i<len) && (i<rec->maxstatcnt); i++) {
    if (map[i] == 0)
      break;

    rec->stat[i].addr = map[i];
    mapcnt++;
  }

  rec->statcnt = mapcnt;
  return mapcnt;
}

int snapshot_rec_alloc_internal(struct snapshot_rec *rec, int maxlinkcnt,
                                int maxstatcnt)
{
  rec->lock = SPIN_LOCK_UNLOCKED;
  rec->packetcnt = 0;
  
  // init local state
  memset(&(rec->local), 0, sizeof(struct snapshot_state));

  // alloc and init link state
  rec->linkcnt = 0;
  rec->maxlinkcnt = maxlinkcnt;
  if (maxlinkcnt != 0) {
    rec->link = vmalloc(sizeof(struct snapshot_state) * maxlinkcnt);
    if (rec->link == NULL) {
      goto err_alloc_link;
    }
    memset(rec->link, 0, sizeof(struct snapshot_state) * maxlinkcnt);
  }

  // alloc and init statistic struct
  rec->statcnt = 0;
  rec->maxstatcnt = maxstatcnt;
  if (maxstatcnt != 0) {
    rec->stat = vmalloc(sizeof(struct snapshot_statistic) * maxstatcnt);
    if (rec->stat == NULL) {
      goto err_alloc_stat;
    }
    memset(rec->stat, 0, sizeof(struct snapshot_state) * maxstatcnt);
  }

  goto end;
  
err_alloc_stat:
  vfree(rec->link);  
  rec->link = NULL;
  rec->linkcnt = 0;
  rec->maxlinkcnt = 0;
  
err_alloc_link:
  return -ENOMEM;
  
end:
  return 0;
}

void snapshot_rec_free_internal(struct snapshot_rec *rec)
{
  if (rec->link != NULL)
    vfree(rec->link);  
  rec->link = NULL;
  rec->linkcnt = 0;
  rec->maxlinkcnt = 0;

  if (rec->stat != NULL)
    vfree(rec->stat);
  rec->stat = NULL;
  rec->statcnt = 0;
  rec->maxstatcnt = 0;
}

inline int snapshot_find_link_by_dev_addr(struct snapshot_rec *rec,
                                          unsigned char *dev_addr)
{
  int i;
  for_each_link(rec, i) {
    if (memcmp(rec->link[i].dev_addr, dev_addr, rec->link[i].dev_addrlen) == 0)
      return i;
  }
  return -1;
}

inline int snapshot_find_link_by_addr(struct snapshot_rec *rec, __u32 addr)
{
  int i;
  for_each_link(rec, i) {
    if (rec->link[i].addr == addr)
      return i;
  }
  return -1;
}

inline int snapshot_find_stat_by_addr(struct snapshot_rec *rec, __u32 addr)
{
  int i;
  for_each_stat(rec, i) {
    if (rec->stat[i].addr == addr)
      return i;
  }
  return -1;
}

inline int snapshot_print_state(char *page, struct snapshot_state *s)
{
  int len = 0;
  int i;

  char str[17];

  sprintf(str, "%u.%u.%u.%u", NIPQUAD(s->addr));
  str[sizeof(str)-1] = '\0';
  
  len += sprintf(page+len, "%-16.16s  ", str);
  
  for (i=0; i<MAX_ADDR_LEN; i++) {
    if (i < s->dev_addrlen) {
      len += sprintf(page+len, "%2.2x%c", s->dev_addr[i],
                     (i==s->dev_addrlen-1) ? ' ' : ':');
    } else {
      len += sprintf(page+len, "   ");
    }
  }
  
  len += sprintf(page+len, " %-13.13s  %u\n",
                 snapshot_state_desc[s->state],
                 s->cnt);
  
  return len;
}

inline int snapshot_print_statistic(char *page, struct snapshot_statistic *s)
{
  int len = 0;
  char str[17];

  sprintf(str, "%u.%u.%u.%u", NIPQUAD(s->addr));
  str[sizeof(str)-1] = '\0';
  
  len += sprintf(page+len, "%-16.16s  %u\n", str, s->cnt);
  
  return len;
}

inline int snapshot_print_rec(char *page, struct snapshot_rec *rec)
{
  int len = 0;
  int i;

  len += sprintf(page+len, "seqno: %u  packet count: %u   link count: %d   finished: %d\n",
                 rec->seqno, rec->packetcnt, rec->linkcnt, rec->finished);

  len += sprintf(page+len, "local state\n");
  len += snapshot_print_state(page+len, &(rec->local));

  if (rec->linkcnt != 0) {
    len += sprintf(page+len, "link states\n");
    for_each_link(rec, i) {
    len += snapshot_print_state(page+len, &(rec->link[i]));
    }
  }

  if (rec->statcnt != 0) {
    len += sprintf(page+len, "statistics\n");
    for_each_stat(rec, i) {
      len += snapshot_print_statistic(page+len, &(rec->stat[i]));
    }
  }
  
  return len;
}

void snapshot_clear(struct snapshot_rec *rec)
{
  int i;
  
  rec->finished = 0;
  rec->seqno = 0;
  rec->packetcnt = 0;

  rec->local.cnt = 0;
  rec->local.state = SNAPSHOT_STATE_READY;

  for_each_link(rec, i) {
    rec->link[i].cnt = 0;
    rec->link[i].state = SNAPSHOT_STATE_READY;
  }

  for_each_stat(rec, i) {
    rec->stat[i].cnt = 0;
  }
}

void snapshot_reset(struct snapshot_rec *rec)
{
  rec->victim_addr = 0;
  rec->snapshotcnt = 0;
    
  rec->finished = 0;
  rec->seqno = 0;
  rec->packetcnt = 0;

  memset(&(rec->local), 0, sizeof(struct snapshot_state)); 

  rec->linkcnt = 0;

  rec->statcnt = 0;
}
