/*
 *  changeDetect.c
 *
 *  The file contain function that do the change detection.
 *
 */

#include <lib/util.h>

#define MAX(a, b) (a)>(b)? (a):(b)

/*
 *  Change detection:
 *
 *  Using The CUSUM RLS filter 
 *  (Ref. P.69, Adaptive Filtering and Change Detection, Fredrik Gustafsson)
 *
 *  Algo:
 *  
 *  newMean = a*pastMean + (1-a)*newSample;   // exp_m
 *  e = newSample - pastMean;
 *  s1 = e;
 *  s2 = -e;
 *  g1t = max(g1t+s1-v, 0);
 *  g2t = max(g2t+s2-v, 0);
 *  
 *  if (g1t > h || g2t > h)
 *  {
 *      g1t = 0;                    // reset 
 *      g2t = 0;
 *      pastMean = newSample;
 *      
 *      return Alarm;
 *  }
 *  pastMean = newMean;
 *
 *
 *  How to tune the filter:
 *
 *  Start with a very large threshold h. Choose v to one half of the 
 *  expected change, or adject v such that gt = 0 more than 50% of the
 *  time. Then set the threshold so the required number of false 
 *  alarms (this can be done automatically) or delay for detection is 
 *  obtained.
 *  - If faster detection is sought, try to decrease v.
 *  - If fewer false alarms are wanted, try to increse v.
 *  - If there is a subset of the change times that does not make
 *    sense, try to increse v.
 * 
 */
typedef struct changeDetectStruct 
{
    double pastMean;    // past data of samples
    double g1t;         // test statistic g
    double g2t;         // test statistic g
    double v;           // forgetting parameter
    double h;           // the threadhold
    double a;           // the alpha for exp mean
    
} ChangeDetectStruct;

typedef struct calibrateStruct
{
    int     maxNumOfSample;
    int     numOfSample;
    double *samplesArray;
} CalibrateStruct;


// exp_m
// return (pastmean*a + newSample*(1-a))
double exp_m(double pastmean, double newSample, double a)
{
    if (a < 0 || a > 1)
        return 0;
    return (pastmean*a + newSample*(1-a));
}

ChangeDetectStruct* NewChangeDetectStruct(double v, double h)
{
    ChangeDetectStruct* s = (ChangeDetectStruct*)malloc(sizeof(ChangeDetectStruct));
    s->pastMean = 0;
    s->g1t = 0;
    s->v = v;
    s->h = h;
    return s;
}

void ResetChangeDetectStruct(ChangeDetectStruct* s, double resetMean)
{
    s->g1t = 0;
    s->g2t = 0;
    s->pastMean = resetMean;
}

double ChangeDetect(ChangeDetectStruct* s, double newSample, 
                    double* returnG1t, double* returnG2t )
{
    double newMean;
    double e, s1, s2, returnValue;
    
    returnValue = 0;
    newMean = exp_m(s->pastMean, newSample, s->a);
    e = newSample - s->pastMean;
    s1 = e;
    s2 = -e;
    s->g1t = MAX(s->g1t + s1 - s->v, 0);
    s->g2t = MAX(s->g2t + s2 - s->v, 0);
    
    s->pastMean = newMean;
    *returnG1t = s->g1t;
    *returnG2t = s->g2t;
    DB("g1t: %lf\tg2t: %lf\n", s->g1t, s->g2t);
    if (s->g1t > s->h)      // detect 'up' side load change
        returnValue = 1;
    if (s->g1t > s->h)      // detect 'down' side load change
        returnValue = 2;
    if (s->g1t > s->h || s->g2t > s->h)
        ResetChangeDetectStruct(s, newSample);
    return returnValue;
}

CalibrateStruct* NewCalibrateStruct(int maxNumOfSample)
{
    CalibrateStruct* cs = (CalibrateStruct*)malloc(sizeof(CalibrateStruct));
    cs->maxNumOfSample = maxNumOfSample;
    cs->numOfSample = 0;
    cs->samplesArray = (double*)malloc(maxNumOfSample*sizeof(double));
    return cs;
}

void FreeCalibrateStruct(CalibrateStruct* cs)
{
    free(cs->samplesArray);
    free(cs);
}


int AddCalibrateSample(CalibrateStruct* cs, double sample)
{   
    if (cs->numOfSample < cs->maxNumOfSample)
    {
        cs->samplesArray[cs->numOfSample] = sample;
        cs->numOfSample++;
        return 1;
    }
    return 0;
}

/*
 *  Calibrate
 *
 *  Input: CalibrateStruct* cs: input a CalibrateStruct that have samplies.
 *         reqNumOfFalseAlarm:  The max false alarm allowed in the change
 *                              detection of samples. 
 *         expectedChange:      The expected change of the sudden load
 *                              increase, or -1 so the algorithm will detect 
 *                              by itselves.                              
 */
ChangeDetectStruct* Calibrate(CalibrateStruct* cs, int reqNumOfFalseAlarm, int expectedChange)
{
    ChangeDetectStruct* cds;
    double v, h, mean, max;
    double delta;
    int i;
    int num;
    double* g1t = (double*)malloc(cs->numOfSample*sizeof(double));
    
    cds = (ChangeDetectStruct*)malloc(sizeof(ChangeDetectStruct));

    mean = 0;
    for (i=0; i<cs->numOfSample; i++)
    {
        mean += cs->samplesArray[i];
    }
    mean = mean/cs->numOfSample;
    v = mean;
    delta = v/2;

    if (expectedChange != -1)
    {
        max = 0;
        v = expectedChange/2;
        g1t[0] = 0;
        for (i=1; i<cs->numOfSample; i++)
        {
            g1t[i] = MAX(g1t[i-1] + cs->samplesArray[i] - v, 0);
            max = MAX(g1t[i], max);
        }
    }
    else
    {
        // use binary method to try the value of v, stop if delta is too small
        while (1)
        {
            g1t[0] = 0;
            num = 0;
            max = 0;
            for (i=1; i<cs->numOfSample; i++)
            {
                g1t[i] = MAX(g1t[i-1] + cs->samplesArray[i] - v, 0);
                max = MAX(g1t[i], max);
                if (g1t[i] == 0)
                    num++;
            }
            if (num == cs->numOfSample/2 || delta < v/1000)
                break;
            if (num > cs->numOfSample/2)
                v -= delta;
            else
                v += delta;
            delta /= 2;
        }
    }
    
    // use binary method to try the value of h, stop if delta is too small
    num = 0;
    h = max;
    delta = h/2;
    while (1)
    {
        for (i=0; i<cs->numOfSample; i++)
        {
            if (g1t[i] > h)
                num++;
        }
        if (num == reqNumOfFalseAlarm || delta < h/1000)
            break;
        delta /= 2;
        if (num < reqNumOfFalseAlarm)
            h -= delta;
        else
            h += delta;
        DB("num: %d, v: %lf, h: %lf\n", num, v, h);
    }       
    cds->v = v;
    cds->h = h;
    cds->pastMean = mean;
    cds->g1t = 0;
    cds->g2t = 0;
    cds->a = 0.5;
    DB("v: %lf, h: %lf\n", v, h);
    return cds;
}
